from typing import List

import numpy as np
from pymatgen.core import Structure

from dsenum.utils import get_symmetry_operations


def generate_all_superlattices(index: int) -> List[np.ndarray]:
    """
    enumerate 3-by-3 Hermite normal form which determinant is equal to `index`

    Parameters
    ----------
    index: positive integer

    Returns
    -------
    list_HNF: list of matrices, each element is Hermite normal form
    """

    def make_HNF(a, b, c, d, e, f):
        arr = np.zeros((3, 3), dtype=int)
        arr[np.tril_indices(3)] = np.array([a, b, c, d, e, f])
        return arr

    list_HNF = []
    for a in range(1, index + 1):
        if index % a != 0:
            continue
        for c in range(1, index // a + 1):
            if (index % c != 0) or (index % (a * c) != 0):
                continue
            f = index // (a * c)
            list_HNF.extend(
                [make_HNF(a, b, c, d, e, f) for b in range(c) for d in range(f) for e in range(f)]
            )

    return list_HNF


def reduce_HNF_list_by_parent_lattice_symmetry(
    list_HNF: List[np.ndarray], list_rotation_matrix: np.ndarray
) -> List[np.ndarray]:
    """
    reduce equivalent HNF with parent lattice symmetry

    Parameters
    ----------
    list_HNF: list of matrices, each element is Hermite normal form
    list_rotation_matrix: list of matrices
        each element represents the symmetry of parent lattice
    lattice_vector: array, (3, 3)
        lattice_vector[:, i] is the i-th lattice vector.

    Returns
    -------
    list_reduced_HNF: list of matrices, unique by symmetry
    """

    def is_equivalent(Bi, list_RBj_inv):
        if list_RBj_inv is None:
            return False
        list_H = np.dot(list_RBj_inv, Bi)
        close = np.isclose(list_H, np.around(list_H))
        if np.any(np.all(close, axis=(1, 2)), axis=-1):
            return True
        return False

    list_reduced_HNF = []
    list_RBj_inv = None

    sgn = np.linalg.det(list_rotation_matrix) == 1
    rotations = list_rotation_matrix[sgn, ...]

    for Bi in list_HNF:
        if not is_equivalent(Bi, list_RBj_inv):
            list_reduced_HNF.append(Bi)
            preinv = np.linalg.solve(np.dot(rotations, Bi), np.identity(Bi.shape[0])[None, ...])
            if list_RBj_inv is None:
                list_RBj_inv = preinv
            else:
                list_RBj_inv = np.concatenate([list_RBj_inv, preinv])

    return list_reduced_HNF


def generate_symmetry_distinct_superlattices(
    index: int,
    structure: Structure,
    return_symops=False,
    symprec=1e-2,
):
    """
    generate symmetry distict HNF

    Parameters
    ----------
    index: positive integer
    structure: pymatgen.core.Structure
    return_symops: bool

    Returns
    -------
    list_reduced_HNF: list of matrices, unique by symmetry
    (Optional) rotations: array, (# of symmetry operations, 3, 3)
    (Optional) translations: array, (# of symmetry operations, 3)
    """
    rotations, translations = get_symmetry_operations(structure, symprec=symprec)
    list_HNF = generate_all_superlattices(index)
    list_reduced_HNF = reduce_HNF_list_by_parent_lattice_symmetry(list_HNF, rotations)
    if return_symops:
        return list_reduced_HNF, rotations, translations
    else:
        return list_reduced_HNF



import pytest
from pymatgen.analysis.structure_matcher import StructureMatcher
from tqdm import tqdm

from dsenum.coloring import SiteColoringEnumerator
from dsenum.coloring_generator import (
    ColoringGenerator,
    FixedConcentrationColoringGenerator,
)
from dsenum.enumerate import StructureEnumerator, ZddStructureEnumerator
from dsenum.permutation_group import DerivativeStructurePermutation
from dsenum.polya import polya_counting, polya_fixed_degrees_counting
from dsenum.superlattice import generate_symmetry_distinct_superlattices
from dsenum.utils import get_lattice

FCC_BINARY_COUNTS = [
    2,
    3,
    12,
    14,
    50,
    52,
    229,
    252,
    685,
    682,
    3875,
    2624,
    9628,
    16584,
    49764,
    42135,
    212612,
    174104,
    867893,
    1120708,
    2628180,
    3042732,
]


@pytest.mark.parametrize(
    "kind,num_types,index_start,num_expected",
    [
        ("fcc", 2, 2, FCC_BINARY_COUNTS),
        ("sc", 2, 2, [3, 3, 15]),
        ("fcc", 3, 3, [3, 13, 23, 130, 197, 1267, 2322, 9332]),
        ("fcc", 4, 4, [7, 9, 110, 211, 2110, 5471, 32362]),
        ("hcp", 2, 2, [7, 30, 163, 366, 2613, 5268, 42901, 119528, 662193]),
    ],
)
@pytest.mark.parametrize("method", ["direct", "lexicographic"])
def test_colorings_small(kind, num_types, index_start, num_expected, method):
    # Skip too time-consuming cases...
    max_index = 7

    structure = get_lattice(kind)
    index_end = index_start + len(num_expected)
    for index, expected in zip(range(index_start, index_end), num_expected):
        if index > max_index:
            continue
        se = StructureEnumerator(
            structure,
            index,
            num_types,
            color_exchange=True,
            remove_superperiodic=True,
            method=method,
        )
        actual = se.generate()
        assert len(actual) == expected


@pytest.mark.parametrize(
    "kind,num_types,composition_constraints,site_constraints",
    [
        ("fcc", 2, None, None),
        ("sc", 2, None, None),
        ("fcc", 3, None, None),
        ("fcc", 4, None, None),
        ("hcp", 2, None, None),
        ("fcc", 2, [2, 1], None),
        ("fcc", 3, [1, 2, 1], None),
        ("hcp", 2, None, [[0], [1]]),
    ],
)
def test_zdd_with_naive(kind, num_types, composition_constraints, site_constraints):
    structure = get_lattice(kind)

    for index in range(1, 6 + 1):
        if composition_constraints:
            ratio_sum = sum(composition_constraints)
            if structure.num_sites * index % ratio_sum != 0:
                # Incorrect composition ratio
                continue

        se = StructureEnumerator(
            structure,
            index,
            num_types,
            composition_constraints=composition_constraints,
            base_site_constraints=site_constraints,
            color_exchange=False,
            remove_incomplete=False,
            remove_superperiodic=False,
        )
        zse = ZddStructureEnumerator(
            structure,
            index,
            num_types,
            composition_constraints=composition_constraints,
            base_site_constraints=site_constraints,
            remove_superperiodic=False,
            remove_incomplete=False,
        )

        count_naive = len(se.generate())
        count_zdd = len(zse.generate())
        assert count_zdd == count_naive


@pytest.mark.parametrize(
    "kind,num_types,composition_constraints,site_constraints",
    [
        ("fcc", 2, None, None),
        ("sc", 2, None, None),
        ("fcc", 3, None, None),
        ("fcc", 4, None, None),
        ("hcp", 2, None, None),
        ("fcc", 2, [2, 1], None),
        ("fcc", 3, [1, 2, 1], None),
        ("hcp", 2, None, [[0], [1]]),
    ],
)
def test_enumerated_structures(kind, num_types, composition_constraints, site_constraints):
    structure = get_lattice(kind)

    for index in range(1, 5):
        # TODO: when remove_superperiodic = remove_incomplete = False, this test is failed.
        zse = ZddStructureEnumerator(
            structure,
            index,
            num_types,
            composition_constraints=composition_constraints,
            base_site_constraints=site_constraints,
            remove_superperiodic=True,
            remove_incomplete=True,
        )
        list_dstructs = zse.generate()

        stm = StructureMatcher(ltol=1e-4, stol=1e-4)
        grouped = stm.group_structures(list_dstructs)
        assert len(grouped) == len(list_dstructs)


@pytest.mark.parametrize(
    "kind,num_types",
    [
        ("fcc", 2),
        ("sc", 2),
        ("fcc", 3),
        ("fcc", 4),
        ("hcp", 2),
    ],
)
def test_colorings_with_polya(kind, num_types):
    structure = get_lattice(kind)
    displacement_set = structure.frac_coords
    num_sites_base = structure.num_sites

    for index in range(num_types, 5):
        list_reduced_HNF, rotations, translations = generate_symmetry_distinct_superlattices(
            index, structure, return_symops=True
        )
        num_sites = num_sites_base * index
        cl_generator = ColoringGenerator(num_sites, num_types)

        for hnf in tqdm(list_reduced_HNF):
            ds_permutation = DerivativeStructurePermutation(
                hnf, displacement_set, rotations, translations
            )
            sc_enum = SiteColoringEnumerator(
                num_types,
                ds_permutation,
                cl_generator,
                color_exchange=False,
                remove_superperiodic=False,
                remove_incomplete=False,
            )
            colorings = sc_enum.unique_colorings()
            cnt_polya = polya_counting(sc_enum.permutation_group, num_types)
            assert len(colorings) == cnt_polya


@pytest.mark.parametrize(
    "kind,num_types",
    [
        ("fcc", 2),
        ("sc", 2),
        ("fcc", 3),
        ("fcc", 4),
        ("hcp", 2),
    ],
)
def test_fixed_colorings_with_polya(kind, num_types):
    structure = get_lattice(kind)
    displacement_set = structure.frac_coords
    num_sites_base = structure.num_sites

    for index in range(num_types, 5):
        list_reduced_HNF, rotations, translations = generate_symmetry_distinct_superlattices(
            index, structure, return_symops=True
        )
        num_sites = num_sites_base * index
        # TODO: test at more color_ratio cases
        color_ratio = [1] * (num_types - 1) + [
            num_sites - num_types + 1,
        ]
        cl_generator = FixedConcentrationColoringGenerator(num_sites, num_types, color_ratio)

        for hnf in tqdm(list_reduced_HNF):
            ds_permutaion = DerivativeStructurePermutation(
                hnf, displacement_set, rotations, translations
            )
            sc_enum = SiteColoringEnumerator(
                num_types,
                ds_permutaion,
                cl_generator,
                color_exchange=False,
                remove_superperiodic=False,
                remove_incomplete=False,
            )
            colorings = sc_enum.unique_colorings()
            cnt_polya = polya_fixed_degrees_counting(
                sc_enum.permutation_group, num_types, color_ratio
            )
            assert len(colorings) == cnt_polya



from itertools import product

import numpy as np

from dsenum.converter import DerivativeMultiLatticeHash
from dsenum.site import CanonicalSite
from dsenum.utils import cast_integer_matrix


class DerivativeStructurePermutation:
    """
    Permutation Representation of space group of superlattice

    Parameters
    ----------
    hnf: array, (dim, dim)
        Hermite normal form(lower triangular)
    displacement_set: array, (num_site_parent, dim)
        fractinal coordinates in primitive cell of base structure
    rotations: array, (# of symmetry operations, dim, dim)
        rotations with primitive basis for base structure
    translations: array, (# of symmetry operations, dim)
        translations with primitive basis for base structure

    Attributes
    ----------
    dim : int
        dimention of lattice
    index: int
        # of parent multilattice in super lattice
    num_site: int
        # of sites in unit cell of superlattice
    """

    def __init__(
        self,
        hnf: np.ndarray,
        displacement_set: np.ndarray,
        rotations: np.ndarray,
        translations: np.ndarray,
    ):
        self.hnf = hnf
        self.num_sites_base = len(displacement_set)
        # TODO: check each site in displacement_set is in [0, 1)^dim
        self.displacement_set = displacement_set

        self.dhash = DerivativeMultiLatticeHash(self.hnf, self.displacement_set)

        self.rotations, self.translations = self._get_superlattice_invariant_subgroup(
            rotations, translations
        )

        self.list_dsites = self.dhash.get_distinct_derivative_sites_list()
        self.list_csites = self.dhash.get_canonical_sites_list()

        self._prm_t = self._get_translation_permutations()
        self.prm_rigid = self._get_rigid_permutations()

    @property
    def dim(self):
        return self.dhash.dim

    @property
    def index(self):
        return self.dhash.index

    @property
    def num_sites(self):
        return self.dhash.num_sites

    @property
    def prm_t(self):
        return self._prm_t

    def _get_superlattice_invariant_subgroup(
        self, rotations: np.ndarray, translations: np.ndarray
    ):
        valid_rotations = []
        valid_translations = []

        for R, tau in zip(rotations, translations):
            if not is_same_lattice(np.dot(R, self.hnf), self.hnf):
                continue

            valid_rotations.append(R)
            valid_translations.append(tau)

        assert len(rotations) % len(valid_rotations) == 0
        return np.array(valid_rotations), np.array(valid_translations)

    def _get_translation_permutations(self):
        list_permutations = []
        for add_factor in self.dhash.get_all_factors():
            new_list_csites = []
            for csite in self.list_csites:
                new_factor = self.dhash.modulus_factor(
                    np.array(csite.factor) + np.array(add_factor)
                )
                new_csite = CanonicalSite(csite.site_index, new_factor)
                new_list_csites.append(new_csite)

            # permutation represenation
            perm = [self.dhash.ravel_canonical_site(csite) for csite in new_list_csites]
            assert is_permutation(perm)
            list_permutations.append(perm)

        # assume list_permutations[0] is identity
        assert is_identity_permutation(list_permutations[0])

        return list_permutations

    def _get_rigid_permutations(self):
        identity = list(range(self.num_sites))
        list_permutations = [
            identity,
        ]

        for R, tau in zip(self.rotations, self.translations):
            new_list_csites = []
            for dsite in self.list_dsites:
                frac_coord = self.displacement_set[dsite.site_index] + np.array(dsite.jimage)
                acted_frac_coord = np.dot(R, frac_coord) + tau
                new_csite = self.dhash.hash_frac_coords(acted_frac_coord)
                assert new_csite is not None
                new_list_csites.append(new_csite)

            perm = [self.dhash.ravel_canonical_site(csite) for csite in new_list_csites]
            assert is_permutation(perm)
            if perm not in list_permutations:
                list_permutations.append(perm)

        # this set of permutations is not group!
        return list_permutations

    def get_symmetry_operation_permutations(self):
        """
        return permutation representation of the symmetry group of the supercell
        """
        list_permutations = []

        for p1 in self.prm_t:
            for p2 in self.prm_rigid:
                perm = product_permutations(p1, p2)
                assert perm not in list_permutations
                list_permutations.append(perm)

        # assume list_permutations[0] is identity
        assert is_identity_permutation(list_permutations[0])

        return list_permutations


def is_unimodular(M: np.ndarray) -> bool:
    if np.abs(np.around(np.linalg.det(M))) == 1:
        return True
    else:
        return False


def is_same_lattice(H1: np.ndarray, H2: np.ndarray) -> bool:
    M = np.linalg.solve(H1, H2)
    M_int = cast_integer_matrix(M)

    if np.array_equal(np.dot(H1, M_int), H2):
        assert is_unimodular(M_int)
        return True
    else:
        return False


def is_permutation(perm):
    return len(set(perm)) == len(perm)


def is_identity_permutation(perm):
    if all([index == i for i, index in enumerate(perm)]):
        return True
    else:
        return False


def is_permutation_group(list_permutations):
    list_permutations_tuple = [tuple(perm) for perm in list_permutations]
    if len(set(list_permutations_tuple)) != len(list_permutations_tuple):
        print(list_permutations)
        raise ValueError("not unique permutations")

    # identity
    if not any([is_identity_permutation(perm) for perm in list_permutations]):
        raise ValueError("There is not identity")

    # close
    for p1, p2 in product(list_permutations, repeat=2):
        p1p2 = product_permutations(p1, p2)
        if p1p2 not in list_permutations:
            raise ValueError("not closed in product")

    # inverse
    for perm in list_permutations:
        perm_inv = [-1 for _ in range(len(perm))]
        for i, idx in enumerate(perm):
            perm_inv[idx] = i

        if perm_inv not in list_permutations:
            raise ValueError("not contains inverse")

    return True


def product_permutations(p1, p2):
    """
    (p1 p2)(i) = p1(p2(i))
    """
    perm = [p1[p2[i]] for i in range(len(p1))]
    return perm



from dsenum.coloring_generator import (
    ColoringGenerator,
    FixedConcentrationColoringGenerator,
    satisfy_site_constraints,
)


def test_coloring_generator():
    num_elements = 8
    num_color = 3
    cg = ColoringGenerator(num_elements, num_color)

    list_colorings, flags = cg.generate_all_colorings()
    assert len(set(list_colorings)) == num_color**num_elements
    assert len(set(flags.keys())) == num_color**num_elements

    list_colorings2 = list(cg.yield_coloring())
    assert len({tuple(e) for e in list_colorings2}) == num_color**num_elements


def test_coloring_generator_with_site_constraints():
    num_color = 3
    site_constraints = [
        [0],
        [1],
        [2],
        [1, 2],
        [0, 2],
        [0, 1, 2],
    ]
    num_elements = len(site_constraints)
    cg = ColoringGenerator(num_elements, num_color, site_constraints)

    colorings_expect = {
        (0, 1, 2, 1, 0, 0),
        (0, 1, 2, 1, 0, 1),
        (0, 1, 2, 1, 0, 2),
        (0, 1, 2, 1, 2, 0),
        (0, 1, 2, 1, 2, 1),
        (0, 1, 2, 1, 2, 2),
        (0, 1, 2, 2, 0, 0),
        (0, 1, 2, 2, 0, 1),
        (0, 1, 2, 2, 0, 2),
        (0, 1, 2, 2, 2, 0),
        (0, 1, 2, 2, 2, 1),
        (0, 1, 2, 2, 2, 2),
    }
    list_colorings, flags = cg.generate_all_colorings()
    assert {tuple(e) for e in list_colorings} == colorings_expect
    assert len(set(flags.keys())) == len(colorings_expect)

    list_colorings2 = list(cg.yield_coloring())
    assert {tuple(e) for e in list_colorings2} == colorings_expect


def test_fixed_concentration_coloring_generator():
    num_elements = 4
    num_color = 3
    color_ratio = [1, 1, 2]
    fcg = FixedConcentrationColoringGenerator(num_elements, num_color, color_ratio)

    colorings_expect = {
        (0, 1, 2, 2),
        (0, 2, 1, 2),
        (0, 2, 2, 1),
        (1, 0, 2, 2),
        (1, 2, 0, 2),
        (1, 2, 2, 0),
        (2, 0, 1, 2),
        (2, 0, 2, 1),
        (2, 1, 0, 2),
        (2, 1, 2, 0),
        (2, 2, 0, 1),
        (2, 2, 1, 0),
    }

    list_colorings, flags = fcg.generate_all_colorings()
    assert {tuple(e) for e in list_colorings} == colorings_expect
    assert len(set(flags.keys())) == len(colorings_expect)

    list_colorings2 = list(fcg.yield_coloring())
    assert {tuple(e) for e in list_colorings2} == colorings_expect


def test_fixed_concentration_coloring_generator_with_site_constraints():
    num_elements = 4
    num_color = 3
    color_ratio = [1, 1, 2]
    site_constraints = [[2], [1, 2], [0, 1, 2], [0, 1, 2]]
    fcg = FixedConcentrationColoringGenerator(
        num_elements, num_color, color_ratio, site_constraints
    )

    colorings_expect = {(2, 1, 0, 2), (2, 1, 2, 0), (2, 2, 0, 1), (2, 2, 1, 0)}

    list_colorings, flags = fcg.generate_all_colorings()
    assert {tuple(e) for e in list_colorings} == colorings_expect
    assert len(set(flags.keys())) == len(colorings_expect)

    list_colorings2 = list(fcg.yield_coloring())
    assert {tuple(e) for e in list_colorings2} == colorings_expect


def test_satisfy_site_constraints():
    site_constraints = [
        [0],
        [1],
        [0, 1],
        [0, 1, 2],
    ]
    coloring1 = [0, 1, 1, 0]
    coloring2 = [1, 1, 1, 1]
    assert satisfy_site_constraints(site_constraints, coloring1)
    assert not satisfy_site_constraints(site_constraints, coloring2)



import pytest

from dsenum.polya import polya_counting, polya_fixed_degrees_counting


@pytest.fixture
def dihedral4():
    Dihedral4 = [
        [0, 1, 2, 3],  # E
        [3, 0, 1, 2],  # C4
        [2, 3, 0, 1],
        [1, 2, 3, 0],
        [1, 0, 3, 2],  # mirror
        [2, 1, 0, 3],
        [3, 2, 1, 0],
        [0, 3, 2, 1],
    ]
    return Dihedral4


def test_polya_counting(dihedral4):
    num_color = 2
    expected = 6
    actual = polya_counting(dihedral4, num_color)
    assert actual == expected


def test_fixed_degrees(dihedral4):
    num_color = 2

    num_elements = [
        [4, 0],
        [3, 1],
        [2, 2],
        [1, 3],
        [0, 4],
    ]
    expects = [1, 1, 2, 1, 1]

    for num_element_of_each_color, expected in zip(num_elements, expects):
        actual = polya_fixed_degrees_counting(dihedral4, num_color, num_element_of_each_color)
        assert actual == expected



import numpy as np
import pytest

from dsenum.converter import DerivativeMultiLatticeHash


@pytest.mark.parametrize(
    "transformation",
    [
        np.diag((2, 2, 2)),
        np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]),
        np.array([[-1, 1, 1], [1, -1, 1], [-1, -1, 1]]),  # negative determinat
    ],
)
@pytest.mark.parametrize(
    "frac_coords",
    [
        np.array([[0, 0, 0]]),
        np.array([[0, 0, 0], [0.5, 0.5, 0.5]]),
    ],
)
def test_converter(transformation, frac_coords):
    converter = DerivativeMultiLatticeHash(transformation, frac_coords)

    index = converter.index
    num_sites = converter.num_sites

    all_factors = converter.get_all_factors()
    assert len({tuple(f) for f in all_factors}) == index

    all_lattice_points = converter.get_lattice_points()
    assert len({tuple(f) for f in all_lattice_points}) == index

    all_periodic_sites = converter.get_canonical_sites_list()
    assert len(set(all_periodic_sites)) == num_sites

    all_dsites = converter.get_distinct_derivative_sites_list()
    for dsite in all_dsites:
        csite, derivative_jimage = converter.hash_derivative_site(dsite, return_image=True)
        fcoords1 = converter.get_frac_coords(dsite)
        fcoords2 = (
            converter.displacement_set[dsite.site_index]
            + np.dot(converter.left_inv, csite.factor)
            + np.dot(converter.hnf, derivative_jimage)
        )
        assert np.allclose(fcoords1, fcoords2)

    for csite in all_periodic_sites:
        ind = converter.ravel_canonical_site(csite)
        csite2 = converter.unravel_to_canonical_site(ind)
        assert csite2 == csite



from dataclasses import dataclass
from typing import Tuple


@dataclass(unsafe_hash=True)
class DerivativeSite:
    # jimage indicates index of unitcell
    site_index: int
    jimage: Tuple[int, ...]


@dataclass(unsafe_hash=True)
class CanonicalSite:
    # factor indicates index of lattice points with some periodic condition
    site_index: int
    factor: Tuple[int, ...]



from typing import List, Union

import numpy as np
from pymatgen.core import Lattice, Structure
from pymatgen.core.periodic_table import DummySpecie, Element, Specie
from pymatgen.core.sites import PeriodicSite

import dsenum
from dsenum.converter import DerivativeMultiLatticeHash


class ColoringToStructure:
    def __init__(
        self,
        base_structure: Structure,
        dshash: DerivativeMultiLatticeHash,
        mapping_color_to_species: List[Union[str, Element, Specie, DummySpecie]],
        additional_species=None,
        additional_frac_coords=None,
    ):
        """
        Parameters
        ----------
        base_structure:
            Structure with only ordering species
        dshash:
        mapping_color_to_species:
        additional_species: list of pymatgen.core.Species, optional
            species which are nothing to do with ordering
        additional_frac_coords: np.ndarray, optional
            fractional coordinates of species which are nothing to do with ordering
        """
        self.base_structure = base_structure
        self.dshash = dshash
        self.mapping_color_to_species = mapping_color_to_species
        self.num_sites = self.dshash.num_sites

        self.lattice_matrix = np.dot(self.base_matrix.T, self.dshash.hnf).T
        # lattice of derivative structure
        self.lattice = Lattice(self.lattice_matrix)

        self.canonical_derivative_sites = dshash.get_canonical_and_derivative_sites_list()

        list_coords = []
        for _, dsite in self.canonical_derivative_sites:
            coords = dshash.get_frac_coords(dsite)
            cart_coords = np.dot(coords, self.base_matrix)
            list_coords.append(cart_coords)
        self.list_coords = list_coords

        self.precomputed_psites = [
            [
                PeriodicSite(sp, coords, self.lattice, coords_are_cartesian=True)
                for sp in self.mapping_color_to_species
            ]
            for coords in self.list_coords
        ]

        # additional fixed sites
        self.additional_species = additional_species
        self.additional_frac_coords = additional_frac_coords

        self.additional_psites = []
        if self.additional_species is not None:
            lattice_points = self.dshash.get_lattice_points()
            for sp, disp in zip(self.additional_species, self.additional_frac_coords):
                cart_coords = [
                    np.dot(np.array(disp) + np.array(lp), self.base_matrix)
                    for lp in lattice_points
                ]
                self.additional_psites.extend(
                    [
                        PeriodicSite(sp, coords, self.lattice, coords_are_cartesian=True)
                        for coords in cart_coords
                    ]
                )

        # map color to specie str (e.g. 0 -> "Cu")
        self.mapping_color_to_species_str = [
            str(specie) for specie in self.mapping_color_to_species
        ]

        self._precompute_for_poscar_string()

    def _precompute_for_poscar_string(self):
        # precompute comment and lattice vectors for POSCAR output
        self.head_lines = []
        version = dsenum.__version__
        comment = "generated by dsenum " + version
        self.head_lines.append(comment)

        # scale
        self.head_lines.append("1.0")

        # lattice vectors
        self.head_lines.extend(
            [
                str(self.lattice_matrix[i, 0])
                + " "
                + str(self.lattice_matrix[i, 1])
                + " "
                + str(self.lattice_matrix[i, 2])
                for i in range(3)
            ]
        )

        self.precompute_coords_str = [
            [
                str(psite.frac_coords[0])
                + " "
                + str(psite.frac_coords[1])
                + " "
                + str(psite.frac_coords[2])
                for psite in list_psites
            ]
            for list_psites in self.precomputed_psites
        ]

        if self.additional_species is not None:
            assert len(self.additional_species) == len(self.additional_frac_coords)
            self.additional_species_str = [str(specie) for specie in self.additional_species]
        else:
            self.additional_species_str = None

        if self.additional_psites is not None:
            self.additional_coords_str = [
                str(psite.frac_coords[0])
                + " "
                + str(psite.frac_coords[1])
                + " "
                + str(psite.frac_coords[2])
                for psite in self.additional_psites
            ]
        else:
            self.additional_coords_str = None

    @property
    def base_matrix(self):
        return self.base_structure.lattice.matrix

    def convert_to_structure(self, coloring) -> Structure:
        list_psites = [self.precomputed_psites[i][coloring[i]] for i in range(self.num_sites)]
        if self.additional_psites is not None:
            list_psites.extend(self.additional_psites)

        dstruct = Structure.from_sites(list_psites)

        return dstruct

    def convert_to_poscar_string(self, coloring) -> str:
        list_coords_str = [
            self.precompute_coords_str[i][coloring[i]] for i in range(self.num_sites)
        ]
        if self.additional_psites is not None:
            list_coords_str.extend([s for s in self.additional_coords_str])

        list_species = [
            self.mapping_color_to_species_str[coloring[i]] for i in range(self.num_sites)
        ]

        head_lines = self.head_lines[::]
        poscar_str = prepare_poscar_string(head_lines, list_species, list_coords_str)
        return poscar_str


def prepare_poscar_string(
    head_lines: List[str], list_species: List[str], list_coords_str: List[str]
) -> str:
    # ref: https://www.vasp.at/wiki/index.php/POSCAR
    lines = head_lines

    # species
    # TODO: this part is dominant in runtime
    counter = []
    element = ""
    count = 0
    for species_str in list_species:
        if element == "":
            element = species_str
            count += 1
        elif species_str == element:
            count += 1
        else:
            counter.append((element, count))
            element = species_str
            count = 1
    if element != "":
        counter.append((element, count))
    lines.append(" ".join([e for e, _ in counter]))
    lines.append(" ".join([str(c) for _, c in counter]))

    # fractional coords
    lines.append("Direct")
    lines.extend(list_coords_str)

    poscar_str = "\n".join(lines)
    return poscar_str



# https://github.com/pypa/setuptools_scm/#retrieving-package-version-at-runtime
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("dsenum")
except PackageNotFoundError:
    # package is not installed
    pass



import numpy as np
import pytest
from pymatgen.core import Lattice, Structure
from pymatgen.core.periodic_table import DummySpecie

from dsenum.cluster.point_cluster import EquivalentPointClusterGenerator, PointCluster
from dsenum.converter import DerivativeMultiLatticeHash
from dsenum.site import DerivativeSite
from dsenum.utils import get_symmetry_operations, square2d_lattice_symmetry


def test_point_cluster():
    pcl0 = PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 0, 1))])
    pcl1 = PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 0, 1))])
    pcl2 = PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (1, 1, 1))])
    assert pcl0 == pcl1
    assert pcl0 != pcl2
    assert len(pcl0) == 2

    # string representation
    pcl0_str = "PointCluster[site_index=0, jimage=(0, 0, 0);site_index=0, jimage=(0, 0, 1)]"
    assert str(pcl0) == pcl0_str


@pytest.fixture
def epcg_fcc():
    fcc = Structure(Lattice([[0, 1, 1], [1, 0, 1], [1, 1, 0]]), [DummySpecie("X")], [[0, 0, 0]])
    transformation = np.eye(3, dtype=int)

    epcg = EquivalentPointClusterGenerator.from_structure(fcc, transformation)
    return epcg


@pytest.fixture
def epcg_bcc():
    bcc = Structure(Lattice([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]), [DummySpecie("X")], [[0, 0, 0]])
    transformation = np.eye(3, dtype=int)

    epcg = EquivalentPointClusterGenerator.from_structure(bcc, transformation)
    return epcg


@pytest.fixture
def epcg_hcp():
    # TODO
    raise NotImplementedError


@pytest.fixture
def epcg_perovskite_oxygen_substructure():
    lattice = Lattice(np.eye(3))
    # fmt: off
    perovskite = Structure(lattice,
                           ['Cu', 'La', 'O', 'O', 'O'],
                           [
                               [0, 0, 0],        # B=Cu
                               [0.5, 0.5, 0.5],  # A=La
                               [0.5, 0, 0],      # X=O
                               [0, 0.5, 0],      # X=O
                               [0, 0, 0.5],      # X=O
                           ])
    oxygen_substructure = Structure(lattice,
                                    ['O', 'O', 'O'],
                                    [
                                        [0.5, 0, 0],      # X=O
                                        [0, 0.5, 0],      # X=O
                                        [0, 0, 0.5],      # X=O
                                    ])
    # fmt: on

    transformation = np.eye(3, dtype=int)

    frac_coords = oxygen_substructure.frac_coords
    rotations, translations = get_symmetry_operations(perovskite)
    converter = DerivativeMultiLatticeHash(transformation, frac_coords)
    epcg = EquivalentPointClusterGenerator(
        frac_coords=frac_coords,
        rotations=rotations,
        translations=translations,
        converter=converter,
    )
    return epcg


@pytest.fixture
def epcg_square2d():
    frac_coords = np.array([[0, 0]])
    rotations, translations = square2d_lattice_symmetry()
    transformation = np.eye(2, dtype=int)

    converter = DerivativeMultiLatticeHash(transformation, frac_coords)
    epcg = EquivalentPointClusterGenerator(frac_coords, rotations, translations, converter)
    return epcg


def test_normalize_point_cluster(epcg_fcc):
    pcl = PointCluster([DerivativeSite(1, (0, 0, 0)), DerivativeSite(0, (1, 1, 0))])
    normed = PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(1, (-1, -1, 0))])
    assert epcg_fcc.normalize_point_cluster(pcl) == normed


def test_fcc(epcg_fcc):
    distinct_point_clusters = [
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (1, 0, 0))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (1, 1, -1))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 1, 1))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 0, 2))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (2, 1, -1))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (1, 1, 1))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 1, 2))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (2, 2, -2))]),
    ]
    num_equiv_clusters = [
        6,
        3,
        12,
        6,
        12,
        4,
        24,
        3,
    ]

    for pcl, cnt in zip(distinct_point_clusters, num_equiv_clusters):
        equivs = epcg_fcc.find_equivalent_point_clusters(pcl)
        assert len(equivs) == cnt

    fcc = Structure(Lattice([[0, 1, 1], [1, 0, 1], [1, 1, 0]]), [DummySpecie("X")], [[0, 0, 0]])
    grouped_point_clusters, _ = epcg_fcc.get_all_clusters(fcc, cutoff=4.1, order=2)

    assert len(grouped_point_clusters) == len(num_equiv_clusters)
    for equivs, cnt in zip(grouped_point_clusters, num_equiv_clusters):
        assert len(equivs) == cnt

    # TODO: test order >= 3

    # test for finding distinct point clusters within cutoff
    # upto the 8th NN
    distinct_point_clusters_actual = epcg_fcc.get_distinct_point_clusters(
        fcc, cutoff=4.01, order=2
    )
    grouped_point_clusters_actual = [
        epcg_fcc.find_equivalent_point_clusters(pcl) for pcl in distinct_point_clusters_actual
    ]
    for expect, group in zip(num_equiv_clusters, grouped_point_clusters_actual):
        assert len(group) == expect


def test_bcc(epcg_bcc):
    distinct_point_clusters = [
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 0, 1))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 1, 1))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (1, 1, 2))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (1, 2, 2))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (2, 2, 2))]),
        PointCluster([DerivativeSite(0, (0, 0, 0)), DerivativeSite(0, (0, 2, 2))]),
    ]
    num_equiv_clusters = [
        4,
        3,
        6,
        12,
        4,
        3,
    ]

    for pcl, cnt in zip(distinct_point_clusters, num_equiv_clusters):
        equivs = epcg_bcc.find_equivalent_point_clusters(pcl)
        assert len(equivs) == cnt

    bcc = Structure(Lattice([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]), [DummySpecie("X")], [[0, 0, 0]])
    grouped_point_clusters, _ = epcg_bcc.get_all_clusters(bcc, cutoff=4.1, order=2)

    assert len(grouped_point_clusters) == len(num_equiv_clusters)
    for equivs, cnt in zip(grouped_point_clusters, num_equiv_clusters):
        assert len(equivs) == cnt


@pytest.mark.skip(reason="TODO")
def test_hcp(epcg_hcp):
    pass


def test_perovskite_oxygen_substructure(epcg_perovskite_oxygen_substructure):
    lattice = Lattice(np.eye(3))
    # fmt: off
    oxygen_substructure = Structure(lattice,
                                    ['O', 'O', 'O'],
                                    [
                                        [0.5, 0, 0],      # X=O
                                        [0, 0.5, 0],      # X=O
                                        [0, 0, 0.5],      # X=O
                                    ])
    # fmt: on

    # singlet cluster
    grouped_singlets, _ = epcg_perovskite_oxygen_substructure.get_all_clusters(
        oxygen_substructure, cutoff=1e-8, order=1
    )
    assert len(grouped_singlets) == 1
    assert len(grouped_singlets[0]) == 3  # O1, O2, O3 sites are equivalent

    # pair cluster
    grouped_point_clusters, sizes = epcg_perovskite_oxygen_substructure.get_all_clusters(
        oxygen_substructure, cutoff=2.01, order=2
    )
    sizes_expected = [
        np.sqrt(2) / 2,
        1,
        1,
        np.sqrt(6) / 2,
        np.sqrt(2),
        np.sqrt(2),
        np.sqrt(10) / 2,
        np.sqrt(3),
        np.sqrt(14) / 2,
        2,
        2,
    ]
    assert np.allclose(sizes, sizes_expected)


def test_square2d(epcg_square2d):
    distinct_point_clusters = [
        PointCluster([DerivativeSite(0, (0, 0)), DerivativeSite(0, (0, 1))]),
        PointCluster([DerivativeSite(0, (0, 0)), DerivativeSite(0, (1, 1))]),
    ]
    num_equiv_clusters = [
        2,
        2,
    ]

    for pcl, cnt in zip(distinct_point_clusters, num_equiv_clusters):
        equivs = epcg_square2d.find_equivalent_point_clusters(pcl)
        assert len(equivs) == cnt



from pathlib import Path

from setuptools import Extension, setup
from Cython.Build import cythonize

ext_modules = [Extension("dsenum.core", ["src/dsenum/core.pyx"], extra_compile_args=["-O3"])]

# Import the README and use it as the long-description.
with open(Path(__file__).resolve().parent / "README.md") as f:
    long_description = "\n" + f.read()


setup(
    name="dsenum",
    license="MIT",
    description="Derivative structure enumerator for multilattice",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Kohei Shinohara",
    author_email="kohei19950508@gmail.com",
    package_dir={"": "src"},
    package_data={"dsenum": ["py.typed"]},
    python_requires=">=3.8",
    setup_requires=["setuptools_scm", "numpy", "Cython>=0.29.1"],
    install_requires=[
        "setuptools",
        "setuptools_scm",
        "wheel",
        "pymatgen>=2021.3.9",
        "numpy",
        "sympy",  # sympy.utilities.iterables.multiset_permutations
        "scipy",  # scipy.special.binom
        "tqdm",
        "hsnf==0.3.13",
        "pyzdd==0.2.4",
    ],
    extras_require={
        "dev": [
            "ipython",
            "notebook",
            "jupyter_contrib_nbextensions",
            "matplotlib",
            "seaborn",
            "pandas",
            "pytest>=6.0.0",
            "pytest-cov",
            "pytest-benchmark",
            "pre-commit",
            "black",
            "flake8",
            "mypy",
            "isort",
            "pyupgrade",
        ],
        "docs": [
            "sphinx",
            "sphinx-autobuild",
            "nbsphinx==0.8.9",
            "furo",
            "m2r2",
            "ipykernel",
        ],
    },
    tests_require=["pytest"],
    ext_modules=cythonize(ext_modules),
    include_package_data=True,
    zip_safe=False,
    use_scm_version=True,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from time import time
from typing import Literal, cast

import numpy as np
from numpy.typing import NDArray
from pymatgen.core import Structure
from pymatgen.core.periodic_table import DummySpecie
from pymatgen.util.typing import SpeciesLike
from pyzdd import Permutation, Universe
from pyzdd.structure import construct_derivative_structures, enumerate_labelings
from tqdm import tqdm

from dsenum.coloring import SiteColoringEnumerator
from dsenum.coloring_generator import (
    BaseColoringGenerator,
    ColoringGenerator,
    FixedConcentrationColoringGenerator,
    ListBasedColoringGenerator,
)
from dsenum.converter import convert_site_constraints
from dsenum.derivative_structure import ColoringToStructure
from dsenum.permutation_group import DerivativeStructurePermutation
from dsenum.superlattice import generate_symmetry_distinct_superlattices
from dsenum.utils import get_symmetry_operations


class AbstractStructureEnumerator(metaclass=ABCMeta):
    """
    Abstract class for enumerating derivative structures.
    """

    def __init__(
        self,
        base_structure: Structure,
        index: int,
        num_types: int,
        mapping_color_species: list[SpeciesLike] = None,
        composition_constraints=None,
        base_site_constraints=None,
        color_exchange=True,
        remove_superperiodic=True,
        remove_incomplete=True,
        verbose=True,
    ):
        self.base_structure = base_structure
        self.index = index
        self.num_types = num_types
        self.composition_constraints = composition_constraints

        # settings
        self.color_exchange = color_exchange
        self.remove_superperiodic = remove_superperiodic
        self.remove_incomplete = remove_incomplete

        list_reduced_HNF, rotations, translations = generate_symmetry_distinct_superlattices(
            index, base_structure, return_symops=True
        )
        self.list_reduced_HNF = list_reduced_HNF
        self.rotations = rotations
        self.translations = translations

        # self.site_constraints[i] is a list of allowed species at the i-th site in supercell
        self.site_constraints = None
        if base_site_constraints:
            assert len(base_site_constraints) == self.num_sites_base
            self.site_constraints = convert_site_constraints(base_site_constraints, self.index)

        if mapping_color_species and len(mapping_color_species) != self.num_types:
            raise ValueError("mapping_color_species must have num_type species.")
        if mapping_color_species is None:
            mapping_color_species = [DummySpecie(str(i)) for i in range(1, self.num_types + 1)]
        self.mapping_color_species = mapping_color_species

        self.verbose = verbose

    @property
    def num_sites_base(self):
        return self.base_structure.num_sites

    @property
    def num_sites(self):
        return self.num_sites_base * self.index

    def generate(
        self,
        return_colorings: bool = False,
        additional_species: list[SpeciesLike] | None = None,
        additional_frac_coords: NDArray | None = None,
        output: Literal["poscar", "pymatgen"] = "pymatgen",
    ) -> list[Structure | str] | tuple[list[Structure | str], list[NDArray], list[list[int]]]:
        """
        Generate derivative structures

        Parameters
        ----------
        return_colorings: bool, optional
            If true, return transformation matrices and colorings in addition
        additional_species: list[SpeciesLike] | None, optional
            species which are nothing to do with ordering. If specified, append these species in
            returned derivative structures.
        additional_frac_coords: NDArray | None, optional
            fractional coordinates of species which are nothing to do with ordering
        output: Literal['poscar', 'pymatgen']
            In default, `pymatgen.core.Structure` objects are returned. If specified
            `output='poscar'`, POSCAR strings are returned instead.

        Returns
        -------
        list_ds: list[Structure | str]
            List of derivative structures
        list[list[int]]

        list_transformations: list[NDArray], optional
            List of transformation matrices. Returned if `return_colorings=True`
        list_colorings: list[list[int]], optional
            List of colorings. Returned if `return_colorings=True`
        """
        assert (output == "pymatgen") or (output == "poscar")
        start = time()

        displacement_set = self.base_structure.frac_coords
        list_ds = []
        list_transformations = []
        list_colorings = []
        for hnf in tqdm(self.list_reduced_HNF, disable=not self.verbose):
            ds_permutation = DerivativeStructurePermutation(
                hnf, displacement_set, self.rotations, self.translations
            )
            # enumerate colorings
            list_colorings_hnf = self._generate_coloring_with_hnf(
                hnf, ds_permutation, additional_species, additional_frac_coords
            )

            # convert to Structure object
            cts = ColoringToStructure(
                self.base_structure,
                ds_permutation.dhash,
                self.mapping_color_species,
                additional_species=additional_species,
                additional_frac_coords=additional_frac_coords,
            )
            if output == "pymatgen":
                list_ds_hnf = [cts.convert_to_structure(cl) for cl in list_colorings_hnf]
            elif output == "poscar":
                list_ds_hnf = [cts.convert_to_poscar_string(cl) for cl in list_colorings_hnf]

            list_ds.extend(list_ds_hnf)
            if return_colorings:
                list_transformations.extend([hnf] * len(list_ds_hnf))
                list_colorings.extend(list_colorings_hnf)

        end = time()
        if self.verbose:
            print("total: {} (Time: {:.4}sec)".format(len(list_ds), end - start))

        if return_colorings:
            return list_ds, list_transformations, list_colorings
        else:
            return list_ds

    @abstractmethod
    def _generate_coloring_with_hnf(
        self,
        hnf: np.ndarray,
        ds_permutation: DerivativeStructurePermutation,
        additional_species,
        additional_frac_coords,
    ) -> list[list[int]]:
        raise NotImplementedError


class StructureEnumerator(AbstractStructureEnumerator):
    """
    Enumerate derivative structures.

    Parameters
    ----------
    base_structure: pymatgen.core.Structure
        Aristotype for derivative structures
    index: int
        How many times to expand unit cell
    num_types: int
        The number of species in derivative structures.
        `num_types` may be larger than the number of the kinds of species in
        `base_structure`: for example, you consider vacancies in derivative structures.
    mapping_color_species: list[SpecieLike] | None, optional
        If specified, use these species in derivative structures.
        The length of this list should be equal to `num_types`
    composition_constraints: list[int] | None, optional
        composition_constraints[i] is the ratio of the i-th species in mapping_color_species.
        For example, set `composition_constraints = [1, 2]` for enumerating TiO2 structures with
        `mapping_color_species = ["Ti", "O"]`.
    base_site_constraints: list[list[int]] | None, optional
        (num_elements, num_color) e.g. site_constraints[2] = [0, 3, 4] means color of site-2
        in base_structure must be 0, 3, or 4.
    color_exchange: bool, optional
        Iff true, identify color-exchanging
    remove_superperiodic: bool, optional
        Iff true, discard superperiodic coloring
    remove_incomplete: bool, optional
        Iff true, discard structures whose number of types are less then `num_types`.
    method: str, optional
        "direct" or "lexicographic", so far
    n_jobs: int, optional
        core in lexicographic coset enumeration(only used when method='lexicographic')
    verbose: bool, optional
        If true, print progress and number of enumerated structures.
    """

    def __init__(
        self,
        base_structure: Structure,
        index: int,
        num_types: int,
        mapping_color_species: list[SpeciesLike] | None = None,
        composition_constraints: list[int] | None = None,
        base_site_constraints: list[list[int]] | None = None,
        color_exchange: bool = True,
        remove_superperiodic: bool = True,
        remove_incomplete: bool = True,
        method: Literal["direct", "lexicographic"] = "direct",
        n_jobs: int = 1,
        verbose: bool = True,
    ):
        super().__init__(
            base_structure=base_structure,
            index=index,
            num_types=num_types,
            mapping_color_species=mapping_color_species,
            composition_constraints=composition_constraints,
            base_site_constraints=base_site_constraints,
            color_exchange=color_exchange,
            remove_superperiodic=remove_superperiodic,
            remove_incomplete=remove_incomplete,
            verbose=verbose,
        )
        self.method = method
        self.n_jobs = n_jobs

        list_reduced_HNF, rotations, translations = generate_symmetry_distinct_superlattices(
            index, base_structure, return_symops=True
        )
        self.list_reduced_HNF = list_reduced_HNF
        self.rotations = rotations
        self.translations = translations

        # composition constraints
        # typing.cast causes no runtime effect
        if self.composition_constraints is None:
            cl_generator = cast(
                BaseColoringGenerator,
                ColoringGenerator(
                    self.num_sites, self.num_types, site_constraints=self.site_constraints
                ),
            )
        else:
            cl_generator = cast(
                BaseColoringGenerator,
                FixedConcentrationColoringGenerator(
                    self.num_sites,
                    self.num_types,
                    self.composition_constraints,
                    site_constraints=self.site_constraints,
                ),
            )
        self.cl_generator = cl_generator

    def _generate_coloring_with_hnf(
        self,
        hnf: np.ndarray,
        ds_permutation: DerivativeStructurePermutation,
        additional_species,
        additional_frac_coords,
    ) -> list[list[int]]:
        sc_enum = SiteColoringEnumerator(
            self.num_types,
            ds_permutation,
            self.cl_generator,
            self.color_exchange,
            self.remove_superperiodic,
            self.remove_incomplete,
            method=self.method,
            n_jobs=self.n_jobs,
        )
        colorings = sc_enum.unique_colorings()

        return colorings


class ZddStructureEnumerator(AbstractStructureEnumerator):
    """
    Enumerate derivative structures with ZDD acceleration.

    Parameters
    ----------
    base_structure: pymatgen.core.Structure
        Aristotype for derivative structures
    index: int
        How many times to expand unit cell
    num_types: int
        The number of species in derivative structures.
        `num_types` may be larger than the number of the kinds of species
        in `base_structure`: for example, you consider vacancies in derivative structures.
    mapping_color_species: list[int] | None, optional
        If specified, use these species in derivative structures.
        The length of this list should be equal to `num_types`
    composition_constraints: list[int] | None, optional
        composition_constraints[i] is the ratio of the i-th species in mapping_color_species.
        For example, set `composition_constraints = [1, 2]` for enumerating TiO2 structures with
        `mapping_color_species = ["Ti", "O"]`.
    base_site_constraints: list[list[int]] | None, optional
        (num_elements, num_color) e.g. site_constraints[2] = [0, 3, 4] means color of site-2
        in base_structure must be 0, 3, or 4.
    remove_superperiodic: bool, optional
        Iff true, discard superperiodic coloring
    remove_incomplete: bool, optional
        Iff true, discard structures whose number of types are less then `num_types`.
    verbose: bool, optional
        If true, print progress and number of enumerated structures.
    """

    def __init__(
        self,
        base_structure: Structure,
        index: int,
        num_types: int,
        mapping_color_species: list[SpeciesLike] | None = None,
        composition_constraints: list[int] | None = None,
        base_site_constraints: list[list[int]] | None = None,
        remove_superperiodic: bool = True,
        remove_incomplete: bool = True,
        verbose: bool = True,
    ):
        super().__init__(
            base_structure=base_structure,
            index=index,
            num_types=num_types,
            mapping_color_species=mapping_color_species,
            composition_constraints=composition_constraints,
            base_site_constraints=base_site_constraints,
            remove_superperiodic=remove_superperiodic,
            remove_incomplete=remove_incomplete,
            verbose=verbose,
        )

        self.prohibited_site_constraints = []
        if self.site_constraints is not None:
            all_species: set[int] = set()
            all_species.update(range(self.num_types))
            for allowed_species in self.site_constraints:
                prohibited = list(all_species.difference(set(allowed_species)))
                self.prohibited_site_constraints.append(prohibited)
            assert len(self.prohibited_site_constraints) == self.num_sites

    def _generate_coloring_with_hnf(
        self,
        hnf: np.ndarray,
        ds_permutation: DerivativeStructurePermutation,
        additional_species,
        additional_frac_coords,
    ) -> list[list[int]]:
        dd = Universe()

        num_sites = ds_permutation.num_sites
        automorphism = [
            Permutation(sigma) for sigma in ds_permutation.get_symmetry_operation_permutations()
        ]
        translations = [Permutation(sigma) for sigma in ds_permutation._prm_t]

        composition_constraints_dd: list[int] = []
        if self.composition_constraints is not None:
            ratio_sum = np.sum(self.composition_constraints)
            if num_sites % ratio_sum != 0:
                # impossible to satisfy composition constraints
                return []
            else:
                composition_constraints_dd = [
                    ratio * (num_sites // ratio_sum) for ratio in self.composition_constraints
                ]

        construct_derivative_structures(
            dd,
            num_sites=num_sites,
            num_types=self.num_types,
            automorphism=automorphism,
            translations=translations,
            composition_constraints=composition_constraints_dd,
            site_constraints=self.prohibited_site_constraints,
            remove_incomplete=self.remove_incomplete,
            remove_superperiodic=self.remove_superperiodic,
        )

        colorings = list(enumerate_labelings(dd, num_sites, self.num_types))
        return colorings

    def count(self) -> int:
        """
        Count the number of derivative structures
        """
        start = time()

        displacement_set = self.base_structure.frac_coords
        count = 0
        for hnf in tqdm(self.list_reduced_HNF, disable=not self.verbose):
            ds_permutation = DerivativeStructurePermutation(
                hnf, displacement_set, self.rotations, self.translations
            )
            # enumerate colorings
            count_hnf = self._count_with_hnf(hnf, ds_permutation)

            count += count_hnf

        end = time()
        if self.verbose:
            print("total: {} (Time: {:.4}sec)".format(count, end - start))

        return count

    def _count_with_hnf(
        self,
        hnf: np.ndarray,
        ds_permutation: DerivativeStructurePermutation,
    ) -> int:
        dd = Universe()

        num_sites = ds_permutation.num_sites
        automorphism = [
            Permutation(sigma) for sigma in ds_permutation.get_symmetry_operation_permutations()
        ]
        translations = [Permutation(sigma) for sigma in ds_permutation._prm_t]

        composition_constraints_dd: list[int] = []
        if self.composition_constraints is not None:
            ratio_sum = np.sum(self.composition_constraints)
            if num_sites % ratio_sum != 0:
                # impossible to satisfy composition constraints
                return 0
            else:
                composition_constraints_dd = [
                    ratio * (num_sites // ratio_sum) for ratio in self.composition_constraints
                ]

        construct_derivative_structures(
            dd,
            num_sites=num_sites,
            num_types=self.num_types,
            automorphism=automorphism,
            translations=translations,
            composition_constraints=composition_constraints_dd,
            site_constraints=self.prohibited_site_constraints,
            remove_incomplete=self.remove_incomplete,
            remove_superperiodic=self.remove_superperiodic,
        )

        return int(dd.cardinality())


def remove_symmetry_duplicates(
    base_structure,
    hnf,
    num_type,
    list_colorings,
    color_exchange: bool,
    leave_superperiodic: bool,
    use_all_colors: bool,
):
    displacement_set = base_structure.frac_coords
    rotations, translations = get_symmetry_operations(base_structure)
    cl_generator = ListBasedColoringGenerator(num_type, list_colorings)

    ds_permutation = DerivativeStructurePermutation(hnf, displacement_set, rotations, translations)
    sc_enum = SiteColoringEnumerator(
        num_type, ds_permutation, cl_generator, color_exchange, leave_superperiodic, use_all_colors
    )
    colorings = sc_enum.unique_colorings()
    return colorings


def remove_symmetry_duplicates_from_generator(
    base_structure,
    hnf,
    num_type,
    list_colorings,
    color_exchange: bool,
    leave_superperiodic: bool,
    use_all_colors: bool,
    method="direct",
    n_jobs=1,
):
    displacement_set = base_structure.frac_coords
    rotations, translations = get_symmetry_operations(base_structure)
    cl_generator = ListBasedColoringGenerator(num_type, list_colorings)

    ds_permutation = DerivativeStructurePermutation(hnf, displacement_set, rotations, translations)
    sc_enum = SiteColoringEnumerator(
        num_type,
        ds_permutation,
        cl_generator,
        color_exchange,
        leave_superperiodic,
        use_all_colors,
        method=method,
        n_jobs=n_jobs,
    )

    colorings = sc_enum.unique_colorings()
    return colorings
